@@ -28,10 +28,12 @@ class S256Point:
             self.usec = usec
             self.csec = None
             sec_cache = usec
+            self.parity = usec[-1] & 1
         elif csec:
             self.csec = csec
             self.usec = None
             sec_cache = csec
+            self.parity = csec[0] - 2
         else:
             raise RuntimeError("need a serialization")
         self.c = ffi.new("secp256k1_pubkey *")
@@ -50,13 +52,18 @@ class S256Point:
         coef = coefficient % N
         new_key = ffi.new("secp256k1_pubkey *")
         s = self.sec(compressed=False)
-        lib.secp256k1_ec_pubkey_parse(GLOBAL_CTX, new_key, s, len(s))
-        lib.secp256k1_ec_pubkey_tweak_mul(GLOBAL_CTX, new_key, coef.to_bytes(32, "big"))
+        if not lib.secp256k1_ec_pubkey_parse(GLOBAL_CTX, new_key, s, len(s)):
+            raise RuntimeError("libsecp256k1 parse error")
+        if not lib.secp256k1_ec_pubkey_tweak_mul(
+            GLOBAL_CTX, new_key, int_to_big_endian(coef, 32)
+        ):
+            raise RuntimeError("libsecp256k1 multiplication error")
         serialized = ffi.new("unsigned char [65]")
         output_len = ffi.new("size_t *", 65)
-        lib.secp256k1_ec_pubkey_serialize(
+        if not lib.secp256k1_ec_pubkey_serialize(
             GLOBAL_CTX, serialized, output_len, new_key, lib.SECP256K1_EC_UNCOMPRESSED
-        )
+        ):
+            raise RuntimeError("libsecp256k1 serialization error")
         return self.__class__(usec=bytes(serialized))
 
     def __add__(self, scalar):
@@ -64,20 +71,35 @@ class S256Point:
         coef = scalar % N
         new_key = ffi.new("secp256k1_pubkey *")
         s = self.sec(compressed=False)
-        lib.secp256k1_ec_pubkey_parse(GLOBAL_CTX, new_key, s, len(s))
-        lib.secp256k1_ec_pubkey_tweak_add(GLOBAL_CTX, new_key, coef.to_bytes(32, "big"))
+        if not lib.secp256k1_ec_pubkey_parse(GLOBAL_CTX, new_key, s, len(s)):
+            raise RuntimeError("libsecp256k1 parse error")
+        if not lib.secp256k1_ec_pubkey_tweak_add(
+            GLOBAL_CTX, new_key, int_to_big_endian(coef, 32)
+        ):
+            raise RuntimeError("libsecp256k1 add error")
         serialized = ffi.new("unsigned char [65]")
         output_len = ffi.new("size_t *", 65)
-        lib.secp256k1_ec_pubkey_serialize(
+        if not lib.secp256k1_ec_pubkey_serialize(
             GLOBAL_CTX, serialized, output_len, new_key, lib.SECP256K1_EC_UNCOMPRESSED
-        )
+        ):
+            raise RuntimeError("libsecp256k1 serialize error")
         return self.__class__(usec=bytes(serialized))
 
     def verify(self, z, sig):
-        msg = z.to_bytes(32, "big")
+        msg = int_to_big_endian(z, 32)
         sig_data = sig.cdata()
         return lib.secp256k1_ecdsa_verify(GLOBAL_CTX, sig_data, msg, self.c)
 
+    def verify_schnorr(self, msg, sig):
+        xonly_key = ffi.new("secp256k1_xonly_pubkey *")
+        if not lib.secp256k1_xonly_pubkey_from_pubkey(
+            GLOBAL_CTX, xonly_key, ffi.NULL, self.c
+        ):
+            raise RuntimeError("libsecp256k1 xonly pubkey error")
+        return lib.secp256k1_schnorrsig_verify(
+            GLOBAL_CTX, sig.raw, msg, len(msg), xonly_key
+        )
+
     def sec(self, compressed=True):
         """returns the binary version of the SEC format"""
         if compressed:
@@ -85,13 +107,14 @@ class S256Point:
                 serialized = ffi.new("unsigned char [33]")
                 output_len = ffi.new("size_t *", 33)
 
-                lib.secp256k1_ec_pubkey_serialize(
+                if not lib.secp256k1_ec_pubkey_serialize(
                     GLOBAL_CTX,
                     serialized,
                     output_len,
                     self.c,
                     lib.SECP256K1_EC_COMPRESSED,
-                )
+                ):
+                    raise RuntimeError("libsecp256k1 serialize error")
                 self.csec = bytes(ffi.buffer(serialized, 33))
             return self.csec
         else:
@@ -99,16 +122,29 @@ class S256Point:
                 serialized = ffi.new("unsigned char [65]")
                 output_len = ffi.new("size_t *", 65)
 
-                lib.secp256k1_ec_pubkey_serialize(
+                if not lib.secp256k1_ec_pubkey_serialize(
                     GLOBAL_CTX,
                     serialized,
                     output_len,
                     self.c,
                     lib.SECP256K1_EC_UNCOMPRESSED,
-                )
+                ):
+                    raise RuntimeError("libsecp256k1 serialize error")
                 self.usec = bytes(ffi.buffer(serialized, 65))
             return self.usec
 
+    def bip340(self):
+        # returns the binary version of BIP340 pubkey
+        xonly_key = ffi.new("secp256k1_xonly_pubkey *")
+        if not lib.secp256k1_xonly_pubkey_from_pubkey(
+            GLOBAL_CTX, xonly_key, ffi.NULL, self.c
+        ):
+            raise RuntimeError("libsecp256k1 xonly pubkey error")
+        output32 = ffi.new("unsigned char [32]")
+        if not lib.secp256k1_xonly_pubkey_serialize(GLOBAL_CTX, output32, xonly_key):
+            raise RuntimeError("libsecp256k1 xonly serialize error")
+        return bytes(ffi.buffer(output32, 32))
+
     def hash160(self, compressed=True):
         # get the sec
         sec = self.sec(compressed)
@@ -135,6 +171,20 @@ class S256Point:
         """Returns the RedeemScript for a p2sh-p2wpkh redemption"""
         return self.p2wpkh_script().redeem_script()
 
+    def p2tr_script(self):
+        """Returns the p2tr Script object"""
+        # avoid circular dependency
+        from buidl.taproot import TapRoot
+
+        return TapRoot(self).script_pubkey()
+
+    def p2pk_tap_script(self):
+        """Returns the p2tr Script object"""
+        # avoid circular dependency
+        from buidl.script import P2PKTapScript
+
+        return P2PKTapScript(self)
+
     def address(self, compressed=True, network="mainnet"):
         """Returns the p2pkh address string"""
         return self.p2pkh_script(compressed).address(network)
@@ -147,6 +197,10 @@ class S256Point:
         """Returns the p2sh-p2wpkh base58 address string"""
         return self.p2wpkh_script().p2sh_address(network)
 
+    def p2tr_address(self, network="mainnet"):
+        """Returns the p2tr bech32m address string"""
+        return self.p2tr_script().address(network)
+
     def verify_message(self, message, sig):
         """Verify a message in the form of bytes. Assumes that the z
         is calculated using hash256 interpreted as a big-endian integer"""
@@ -158,12 +212,27 @@ class S256Point:
         return self.verify(z, sig)
 
     @classmethod
-    def parse(self, sec_bin):
+    def parse(cls, binary):
+        """returns a Point object from a SEC or BIP340 pubkey"""
+        if len(binary) == 32:
+            return cls.parse_bip340(binary)
+        elif len(binary) in (33, 65):
+            return cls.parse_sec(binary)
+        else:
+            raise ValueError(f"Unknown public key format {binary.hex()}")
+
+    @classmethod
+    def parse_sec(cls, sec_bin):
         """returns a Point object from a SEC binary (not hex)"""
         if sec_bin[0] == 4:
-            return S256Point(usec=sec_bin)
+            return cls(usec=sec_bin)
         else:
-            return S256Point(csec=sec_bin)
+            return cls(csec=sec_bin)
+
+    @classmethod
+    def parse_bip340(cls, binary):
+        sec_bin = b"\x02" + binary
+        return cls(csec=sec_bin)
 
 
 G = S256Point(
@@ -198,9 +267,10 @@ class Signature:
         if not self.der_cache:
             der = ffi.new("unsigned char[72]")
             der_length = ffi.new("size_t *", 72)
-            lib.secp256k1_ecdsa_signature_serialize_der(
+            if not lib.secp256k1_ecdsa_signature_serialize_der(
                 GLOBAL_CTX, der, der_length, self.c
-            )
+            ):
+                raise RuntimeError("libsecp256k1 der serialize error")
             self.der_cache = bytes(ffi.buffer(der, der_length[0]))
         return self.der_cache
 
@@ -212,6 +282,35 @@ class Signature:
         return cls(der=der)
 
 
+class SchnorrSignature:
+    def __init__(self, raw):
+        self.raw = raw
+        if len(raw) != 64:
+            raise ValueError("signature should be 64 bytes")
+        # check that the sig's R is valid
+        if big_endian_to_int(raw[:32]) == 0:
+            raise AssertionError("R should not be zero")
+        xonly_key = ffi.new("secp256k1_xonly_pubkey *")
+        if not lib.secp256k1_xonly_pubkey_parse(GLOBAL_CTX, xonly_key, raw[:32]):
+            raise ValueError(f"libsecp256k1 invalid R {raw[:32].hex()}")
+        s = big_endian_to_int(raw[32:])
+        if s >= N:
+            raise ValueError(f"{s:x} is greater than or equal to {N:x}")
+
+    def __repr__(self):
+        return f"SchnorrSignature({self.raw[:32].hex()},{self.raw[32:].hex()})"
+
+    def __eq__(self, other):
+        return self.raw == other.raw
+
+    def serialize(self):
+        return self.raw
+
+    @classmethod
+    def parse(cls, raw):
+        return cls(raw)
+
+
 class PrivateKey:
     def __init__(self, secret, network="mainnet", compressed=True):
         self.secret = secret
@@ -223,18 +322,33 @@ class PrivateKey:
         return "{:x}".format(self.secret).zfill(64)
 
     def sign(self, z):
-        secret = self.secret.to_bytes(32, "big")
-        msg = z.to_bytes(32, "big")
+        secret = int_to_big_endian(self.secret, 32)
+        msg = int_to_big_endian(z, 32)
         csig = ffi.new("secp256k1_ecdsa_signature *")
         if not lib.secp256k1_ecdsa_sign(
             GLOBAL_CTX, csig, msg, secret, ffi.NULL, ffi.NULL
         ):
-            raise RuntimeError("something went wrong with c signing")
+            raise RuntimeError("libsecp256k1 ecdsa signing problem")
         sig = Signature(c=csig)
         if not self.point.verify(z, sig):
-            raise RuntimeError("something went wrong with signing")
+            raise RuntimeError("generated signature doesn't verify")
         return sig
 
+    def sign_schnorr(self, msg, aux):
+        if len(msg) != 32:
+            raise ValueError("msg needs to be 32 bytes")
+        if len(aux) != 32:
+            raise ValueError("aux needs to be 32 bytes")
+        keypair = ffi.new("secp256k1_keypair *")
+        if not lib.secp256k1_keypair_create(
+            GLOBAL_CTX, keypair, int_to_big_endian(self.secret, 32)
+        ):
+            raise RuntimeError("libsecp256k1 keypair creation problem")
+        raw_sig = ffi.new("unsigned char [64]")
+        if not lib.secp256k1_schnorrsig_sign(GLOBAL_CTX, raw_sig, msg, keypair, aux):
+            raise RuntimeError("libsecp256k1 schnorr signing problem")
+        return SchnorrSignature(bytes(ffi.buffer(raw_sig, 64)))
+
     def deterministic_k(self, z):
         k = b"\x00" * 32
         v = b"\x01" * 32
@@ -266,6 +380,30 @@ class PrivateKey:
         # sign the message using the self.sign method
         return self.sign(z)
 
+    def wif(self, compressed=True):
+        # convert the secret from integer to a 32-bytes in big endian using int_to_big_endian(num, 32)
+        secret_bytes = int_to_big_endian(self.secret, 32)
+        # prepend b'\xef' on testnet, b'\x80' on mainnet
+        if self.network == "mainnet":
+            prefix = b"\x80"
+        else:
+            prefix = b"\xef"
+        # append b'\x01' if compressed
+        if compressed:
+            suffix = b"\x01"
+        else:
+            suffix = b""
+        # encode_base58_checksum the whole thing
+        return encode_base58_checksum(prefix + secret_bytes + suffix)
+
+    def tweaked(self, tweak):
+        if self.point.parity:
+            s = N - self.secret
+        else:
+            s = self.secret
+        new_secret = (s + tweak) % N
+        return self.__class__(new_secret, network=self.network)
+
     @classmethod
     def parse(cls, wif):
         """Converts WIF to a PrivateKey object"""
@@ -285,19 +423,3 @@ class PrivateKey:
         else:
             raise ValueError("Invalid WIF")
         return cls(secret, network=network, compressed=compressed)
-
-    def wif(self, compressed=True):
-        # convert the secret from integer to a 32-bytes in big endian using num.to_bytes(32, 'big')
-        secret_bytes = self.secret.to_bytes(32, "big")
-        # prepend b'\xef' on testnet, b'\x80' on mainnet
-        if self.network == "mainnet":
-            prefix = b"\x80"
-        else:
-            prefix = b"\xef"
-        # append b'\x01' if compressed
-        if compressed:
-            suffix = b"\x01"
-        else:
-            suffix = b""
-        # encode_base58_checksum the whole thing
-        return encode_base58_checksum(prefix + secret_bytes + suffix)
