@@ -8,8 +8,12 @@ from buidl.helper import (
     encode_base58_checksum,
     hash160,
     hash256,
+    hash_aux,
+    hash_challenge,
+    hash_nonce,
     int_to_big_endian,
     raw_decode_base58,
+    xor_bytes,
 )
 
 
@@ -203,7 +207,10 @@ class S256Field(FieldElement):
         return self.hex()
 
     def sqrt(self):
-        return self ** ((P + 1) // 4)
+        s = self ** ((P + 1) // 4)
+        if s * s != self:
+            raise ValueError(f"{self} does not have a square root in {P:x}")
+        return s
 
 
 class S256Point(Point):
@@ -213,6 +220,12 @@ class S256Point(Point):
             super().__init__(x=S256Field(x), y=S256Field(y), a=a, b=b)
         else:
             super().__init__(x=x, y=y, a=a, b=b)
+        if x is None:
+            return
+        if self.y.num % 2 == 1:
+            self.parity = 1
+        else:
+            self.parity = 0
 
     def __eq__(self, other):
         return self.x == other.x and self.y == other.y
@@ -235,6 +248,9 @@ class S256Point(Point):
         else:
             return super().__add__(other)
 
+    def negate(self):
+        return -1 * self
+
     def sec(self, compressed=True):
         # returns the binary version of the sec format, NOT hex
         # if compressed, starts with b'\x02' if self.y.num is even, b'\x03' if self.y is odd
@@ -242,15 +258,21 @@ class S256Point(Point):
         # remember, you have to convert self.x.num/self.y.num to binary using int_to_big_endian
         x = int_to_big_endian(self.x.num, 32)
         if compressed:
-            if self.y.num % 2 == 0:
-                return b"\x02" + x
-            else:
+            if self.parity:
                 return b"\x03" + x
+            else:
+                return b"\x02" + x
         else:
             # if non-compressed, starts with b'\x04' followod by self.x and then self.y
             y = int_to_big_endian(self.y.num, 32)
             return b"\x04" + x + y
 
+    def bip340(self):
+        # returns the binary version of BIP340 pubkey
+        if self.x is None:
+            return int_to_big_endian(0, 32)
+        return int_to_big_endian(self.x.num, 32)
+
     def hash160(self, compressed=True):
         # get the sec
         sec = self.sec(compressed)
@@ -277,6 +299,20 @@ class S256Point(Point):
         """Returns the RedeemScript for a p2sh-p2wpkh redemption"""
         return self.p2wpkh_script().redeem_script()
 
+    def p2tr_script(self):
+        """Returns the p2tr Script object"""
+        # avoid circular dependency
+        from buidl.taproot import TapRoot
+
+        return TapRoot(self).script_pubkey()
+
+    def p2pk_tap_script(self):
+        """Returns the p2tr Script object"""
+        # avoid circular dependency
+        from buidl.script import P2PKTapScript
+
+        return P2PKTapScript(self)
+
     def address(self, compressed=True, network="mainnet"):
         """Returns the p2pkh address string"""
         return self.p2pkh_script(compressed).address(network)
@@ -289,6 +325,10 @@ class S256Point(Point):
         """Returns the p2sh-p2wpkh base58 address string"""
         return self.p2wpkh_script().p2sh_address(network)
 
+    def p2tr_address(self, network="mainnet"):
+        """Returns the p2tr bech32m address string"""
+        return self.p2tr_script().address(network)
+
     def verify(self, z, sig):
         # remember sig.r and sig.s are the main things we're checking
         # remember 1/s = pow(s, N-2, N)
@@ -311,13 +351,35 @@ class S256Point(Point):
         # verify the message using the self.verify method
         return self.verify(z, sig)
 
+    def verify_schnorr(self, msg, schnorr_sig):
+        if self.y.num % 2 == 1:
+            point = self.negate()
+        else:
+            point = self
+        if schnorr_sig.r.x is None:
+            return False
+        message = schnorr_sig.r.bip340() + point.bip340() + msg
+        e = big_endian_to_int(hash_challenge(message)) % N
+        result = schnorr_sig.s * G + (N - e) * point
+        return result == schnorr_sig.r
+
     @classmethod
-    def parse(self, sec_bin):
-        """returns a Point object from a compressed sec binary (not hex)"""
+    def parse(cls, binary):
+        """returns a Point object from a SEC or BIP340 pubkey"""
+        if len(binary) == 32:
+            return cls.parse_bip340(binary)
+        elif len(binary) in (33, 65):
+            return cls.parse_sec(binary)
+        else:
+            raise ValueError(f"Unknown public key format {binary.hex()}")
+
+    @classmethod
+    def parse_sec(cls, sec_bin):
+        """returns a Point object from a SEC pubkey"""
         if sec_bin[0] == 4:
             x = int(sec_bin[1:33].hex(), 16)
             y = int(sec_bin[33:65].hex(), 16)
-            return S256Point(x=x, y=y)
+            return cls(x=x, y=y)
         is_even = sec_bin[0] == 2
         x = S256Field(int(sec_bin[1:].hex(), 16))
         # right side of the equation y^2 = x^3 + 7
@@ -331,9 +393,25 @@ class S256Point(Point):
             even_beta = S256Field(P - beta.num)
             odd_beta = beta
         if is_even:
-            return S256Point(x, even_beta)
+            return cls(x, even_beta)
         else:
-            return S256Point(x, odd_beta)
+            return cls(x, odd_beta)
+
+    @classmethod
+    def parse_bip340(cls, bip340_bin):
+        """returns a Point object from a BIP340 pubkey"""
+        n = big_endian_to_int(bip340_bin)
+        if n == 0:
+            # point at infinity
+            return cls(None, None)
+        x = S256Field(n)
+        # right side of the equation y^2 = x^3 + 7
+        alpha = x ** 3 + S256Field(B)
+        # solve for left side
+        beta = alpha.sqrt()
+        if beta.num % 2 == 1:
+            beta = S256Field(P - beta.num)
+        return cls(x, beta)
 
 
 G = S256Point(
@@ -398,9 +476,37 @@ class Signature:
         return cls(r, s)
 
 
+class SchnorrSignature:
+    def __init__(self, r, s):
+        self.r = r
+        if s >= N:
+            raise ValueError(f"{s:x} is greater than or equal to {N:x}")
+        self.s = s
+
+    def __repr__(self):
+        return f"SchnorrSignature({self.r},{self.s:x})"
+
+    def __eq__(self, other):
+        return self.r == other.r and self.s == other.s
+
+    def serialize(self):
+        return self.r.bip340() + int_to_big_endian(self.s, 32)
+
+    @classmethod
+    def parse(cls, signature_bin):
+        stream = BytesIO(signature_bin)
+        r = S256Point.parse(stream.read(32))
+        s = big_endian_to_int(stream.read(32))
+        return cls(r, s)
+
+
 class PrivateKey:
     def __init__(self, secret, network="mainnet", compressed=True):
         self.secret = secret
+        if secret > N - 1:
+            raise RuntimeError("secret too big")
+        if secret < 1:
+            raise RuntimeError("secret too small")
         self.point = secret * G
         self.network = network
         self.compressed = compressed
@@ -423,6 +529,29 @@ class PrivateKey:
         # Signature(r, s)
         return Signature(r, s)
 
+    def sign_schnorr(self, msg, aux):
+        if len(msg) != 32:
+            raise ValueError("msg needs to be 32 bytes")
+        if len(aux) != 32:
+            raise ValueError("aux needs to be 32 bytes")
+        if self.point.y.num % 2 == 0:
+            d = self.secret
+        else:
+            d = N - self.secret
+        t = xor_bytes(int_to_big_endian(d, 32), hash_aux(aux))
+        k = big_endian_to_int(hash_nonce(t + self.point.bip340() + msg)) % N
+        r = k * G
+        if r.y.num % 2 == 1:
+            k = N - k
+            r = k * G
+        message = r.bip340() + self.point.bip340() + msg
+        e = big_endian_to_int(hash_challenge(message)) % N
+        s = (k + e * d) % N
+        sig = SchnorrSignature(r, s)
+        if not self.point.verify_schnorr(msg, sig):
+            raise RuntimeError("bad sig")
+        return sig
+
     def deterministic_k(self, z):
         k = b"\x00" * 32
         v = b"\x01" * 32
@@ -470,6 +599,14 @@ class PrivateKey:
         # encode_base58_checksum the whole thing
         return encode_base58_checksum(prefix + secret_bytes + suffix)
 
+    def tweaked(self, tweak):
+        if self.point.parity:
+            s = N - self.secret
+        else:
+            s = self.secret
+        new_secret = (s + tweak) % N
+        return self.__class__(new_secret, network=self.network)
+
     @classmethod
     def parse(cls, wif):
         """Converts WIF to a PrivateKey object"""
